##############################################################################
##
## Install-Zen
##
## by Actian Corporation (http://www.actian.com/zen)
##
##############################################################################

    <#
    .NOTES
        Copyright (c) Actian Corporation.  All rights reserved.

        Use of this sample source code is subject to the terms of the Actian
        Zen license agreement under which you licensed this sample source code.
        If you did not accept the terms of the license agreement, you are not
        authorized to use this sample source code. For the terms of the license,
        please see the license agreement between you and Actian Corporation.
        THE SAMPLE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES.

    .SYNOPSIS
        Installs Zen on a local system running Windows Nano Server or Windows
        IoT Core.

    .DESCRIPTION
        Installs Zen on a local system running either 64-bit Windows Nano Server
        or 32-bit Windows Internet of Things (IoT).  The user must specify one of
        the parameters specified below.  The script can be invoked multiple times
        to update existing files and/or to add new files.

    .PARAMETER enterpriseserver
        Installs Zen Enterprise Server.  On 64-bit Windows Nano Server, the
        64-bit Zen Enterprise Server is installed.  On 32-bit Windows IoT, the 32-bit
        Zen Enterprise Server is installed.

    .PARAMETER client
        Only available on Windows Nano Server.  Installs Zen Client files for
        accessing a Zen database on a remote system.

    .PARAMETER reporting
        Only available on Windows Nano Server.  Installs the 64-bit Zen
        Reporting Engine.

    .EXAMPLE
        PS C:\> .\Install-Zen enterpriseserver
            Install the Zen Enterprise Server.

    .EXAMPLE
        PS C:\> .\Install-Zen client
            Install the Zen Client (Windows Nano Server only).

    .EXAMPLE
        PS C:\> .\Install-Zen reporting
            Installs the Zen Reporting Engine (Windows Nano Server only).

    .INPUTS
        None. You cannot pipe objects to this cmdlet.

    .OUTPUTS
        0 : The script completed with no errors (may contain warnings).
        -1: The script returned one or more errors.

        Creates log file in users $env:temp folder.
        Creates object $ZenInfo with the following properties:
                $ZenInfo.InstallData    : Install location of data files.
                $ZenInfo.InstallDir64   : Install location on 64-bit systems.
                $ZenInfo.InstallDir     : Install location on 32-bit systems,
                                           same as InstallDir64 on 64-bit systems.
                $ZenInfo.IsReport       : workgroup (0) or reporting (1) only
                $ZenInfo.IsVx           : Enterprise Server (0) or Cloud Server (1) only.
                $ZenInfo.ProductVersion : The full Zen major.minor.build version.
                $ZenInfo.VersionLevel   : The Zen major.minor version.

    #>

#Requires -Version 5.1        # -- Require PowerShell 5.1 or higher (Nano Server,IoT).
#Requires -PSEdition Core     # -- Require PowerShell Core
#Requires -RunAsAdministrator # -- Require PowerShell running as Administrator.

param (
    [Parameter(
        Mandatory=$true,
        Position = 0)]
    [ValidateSet('enterpriseserver','client','reporting')]
    [String]
    $InstallProductType
)

Set-StrictMode -Version Latest

Import-LocalizedData msg -FileName "Zen_PowerShell_Script_strings.psd1" -ErrorAction Ignore

#############################
## Variables Required prior to calling SetScriptVariables().
#############################
if ([System.IntPtr]::Size -eq 4) {$script:Is64Bit=$false} else {$script:Is64Bit=$true}
$script:scriptPath = & {Split-Path $myInvocation.ScriptName}
$script:registryPathV13 = "HKLM:\SOFTWARE\Pervasive Software"
$script:registryPath = "HKLM:\SOFTWARE\Actian\Zen"
$script:fpCommonApplicationData=${env:ALLUSERSPROFILE}
if ($Is64Bit) {
    $script:fpProgramFiles64=${env:ProgramFiles}
    $script:fpProgramFiles32=${env:ProgramFiles(x86)}
} else {
    $script:fpProgramFiles32=${env:ProgramFiles}
    $script:fpProgramFiles64=$fpProgramFiles32
}
$script:PsqlDataDir= join-path $fpCommonApplicationData "Actian\PSQL\"
$script:PsqlInstDir64= join-path $fpProgramFiles64 "Actian\PSQL\"
$script:PsqlInstDir32= join-path $fpProgramFiles32 "Actian\PSQL\"
$script:installing=$true
$script:updating=$false
switch ($InstallProductType) {
    'client' { $script:InstallInfoProdType="Client"; break}
    'enterpriseserver' { $script:InstallInfoProdType="enterpriseserver"; break}
    'reporting' { $script:InstallInfoProdType="Workgroup"; break}
}
$script:culture=Get-Culture
if ( $PSCulture.StartsWith("ja") ) {$script:IsJajp=$true} else {$script:IsJajp=$false}

#############################
## Abort Checks Not Using #Requires
#############################
# -- Abort Reporting Engine or Client install if OS is 32-bit.
if (!$Is64Bit -and ($InstallProductType.Equals("reporting") ) ) { throw $msg.errNoRPT32; exit -1 }
# -- Abort if incompatible legacy MIF file detected (enterpriseserver, client, or reporting engine in old location).
if (test-path (join-path $PsqlInstDir64 "bin\w64mif.dll") ) { throw $msg.errNoPSQL; exit -1 }
if (test-path (join-path $PsqlInstDir32 "bin\w3mif190.dll") ) { throw $msg.errNoPSQL; exit -1 }
# -- Abort if other product type detected.
$abortInstall=$False
##  Need to adjust the following to new registry keys: simply match the $installInfoProdType with the
##  child of "HKLM\Software\Actian\InstallInfo"
##
if (test-path $registryPath"\InstallInfo\") {
    # -- Found existing key, verify that any existing sub-keys end with a string matching 
    #    $InstallInfoProdType (see SetScriptVariables).
    #    Assumption #1: For server editions, we allow install of Enterprise Server if Cloud Server
    #                   already installed (and vice versa) so no need to check InstallInfo\IsVx value.
    #    Assumption #2: Workgroup and Reporting are already mutually exclusive based on the OS
    #                   bit-ness checks above so no need to check InstallInfo\IsReport value.
    $alreadyInstalled=""
    Get-ChildItem $registryPath"\InstallInfo\" | ForEach-Object -Process {
        if ($_.PSChildName -eq $installInfoProdType) {
            $updating=$true
        } else {
            $abortInstall=$true
            $alreadyInstalled=$_.PSChildName
        }
    }
}
if ($abortInstall) { throw $msg.errNoOtherPsql -f $InstallProductType, $alreadyInstalled; exit -1}
# -- Abort if install_$InstallProductType not found.
if (Test-Path -Path $scriptPath"\install_"$InstallProductType".cfg") {
} else {
    throw $msg.errNoProductType -f "$scriptPath\install_$InstallProductType.cfg"; exit -1
}
# -- Abort if ProgramFiles ZIP in package doesn't match this platform.
if (Test-Path -Path "$scriptPath\ProgramFilesX64.zip") {$script:expectedPlatform="x86_64"}
if (Test-Path -Path "$scriptPath\ProgramFilesX86.zip") {$script:expectedPlatform="x86"}
if (Test-Path -Path "$scriptPath\ProgramFilesArm32.zip") {$script:expectedPlatform="32-bit ARM"}
if ($env:PROCESSOR_ARCHITECTURE -eq "ARM") {
    # ARM processor, only 32-bit supported.
    if ($Is64Bit) { throw $msg.errNoArm64; exit -1 } else {
        if ($expectedPlatform -ne "32-bit ARM") { throw $msg.errWrongPlatform -f $expectedPlatform,"32-bit ARM"; exit -1}
    }
} else {
    if ($Is64Bit) {
        if ($expectedPlatform -ne "x86_64") { throw $msg.errWrongPlatform -f $expectedPlatform,"x86_64"; exit -1}
    } else {
        if ($expectedPlatform -ne "x86") { throw $msg.errWrongPlatform -f $expectedPlatform,"x86"; exit -1}
    }
}
#############################
## FUNCTIONS
#############################
function Main {
    ##########################
    # Defines the processing order of this script.
    ##########################
    SetScriptVariables
    BeginLog
    ReadFromConfig
    if ($updating) {UnloadPerfMonCounterSets}
    StopPsqlServices
    if ($updating) {RemoveObsoleteFilesOnUpgrade}
    UnpackFiles
    RemoveFilesByProductType
    UpdatePath
    InitPsqlRegistry
    RegisterComInterfaces
    CreatePsqlServices
    AddPsqlFirewallRules
    StartPsqlServices
    ActivatePsqlTempLicense
    CreatePsqlDatabases
    UpdateRoyaltyRecord
    InvokeECollect
    LoadPerfMonCounterSets
    EndLog
}
function SetScriptVariables {
    ##########################
    # Sets script-level variables based on specified product type to be installed. Except
    # for $ZenInstallCfg, of the variable is set to a simple string then it can be
    # overridden by specifying the name=value pair in the install configuration file
    # specified by $ZenInstallCfg.
    #
    # Note: Must be called before logging is started.
    ##########################
    $script:ZenInstallCfg="install_"+ $InstallProductType + ".cfg"
    $script:defaultAclArgs="Everyone","FullControl","ContainerInherit","InheritOnly","Allow"
    $script:scriptName = & {$myInvocation.ScriptName}
    #$script:scriptPath = & {Split-Path $myInvocation.ScriptName}  ## - set before Abort Checks section.
    $script:computerName = ${env:computerName}
    [string]$fwpDomainState=netsh advfirewall show domainprofile State
    if ( $fwpDomainState.Contains("ON") ) {$script:fwDomainProfileState="ON"} else {$script:fwDomainProfileState="OFF"}
    [string]$fwpPrivateState=netsh advfirewall show privateprofile State
    if ( $fwpPrivateState.Contains("ON") ) {$script:fwPrivateProfileState="ON"} else {$script:fwPrivateProfileState="OFF"}
    [string]$fwpPublicState=netsh advfirewall show publicprofile State
    if ( $fwpPublicState.Contains("ON") ) {$script:fwPublicProfileState="ON"} else {$script:fwPublicProfileState="OFF"}
    #$script:registryPath = "HKLM:\SOFTWARE\Actian"  ## - Set before Abort Checks section.
    $script:CmnSvcDispNameOld = "Actian PSQL"
    $script:CmnSvcDispName = "Actian Zen"
    # -- Default ELS settings.
    $script:zenElsUrl=""
    $script:zenElsServerID="0"
    # -- Duplicating MSI behavior, folder paths end with a backslash.
    # -- The following were moved to before Abort Checks section so they could be referenced earlier.
    #$script:fpCommonApplicationData=${env:ALLUSERSPROFILE}
    #if ($Is64Bit) {
    #    $script:fpProgramFiles64=${env:ProgramFiles}
    #    $script:fpProgramFiles32=${env:ProgramFiles(x86)}
    #} else {
    #    $script:fpProgramFiles32=${env:ProgramFiles}
    #    $script:fpProgramFiles64=$fpProgramFiles32
    #}
    #$script:PsqlDataDir= join-path $fpCommonApplicationData "Actian\PSQL\"
    #$script:PsqlInstDir64= join-path $fpProgramFiles64 "Actian\PSQL\"
    #$script:PsqlInstDir32= join-path $fpProgramFiles32 "Actian\PSQL\" 
    $script:ZenDataDir= join-path $fpCommonApplicationData "Actian\Zen\"
    $script:ZenInstDir64= join-path $fpProgramFiles64 "Actian\Zen\"
    $script:ZenInstDir32= join-path $fpProgramFiles32 "Actian\Zen\"
    $script:setupdbCfgInstall= join-path $ZenDataDir "w32setupdb.cfg"
    # -- Default values (overwritten based on $InstallProductType)
    $script:isClient=$False
    $script:isServer=$False
    $script:isReporting=$False
    $script:installLogFile = join-path $env:TEMP "\Zen_PowerShell_Install.log"
    # -- Variables based on command line arugments.
    switch ($InstallProductType) {
        'client' {
            $script:isClient=$True
            $script:InstallInfoProdType="Client"
            $script:installLogFile = join-path $env:TEMP "Zen_Client_PowerShell_Install.log"
            $script:filesToRemoveFromBin = "zenengnsvc32.exe", "uwmkde.dll","w64repmkde.dll","ntmkde.dll","w64ntmkde.dll","psqlperfctr*.xml"
            $script:filesToRemoveFromZen = ""
            $script:filesToRemoveFromData = "collate.cfg","defaultdb","Demodata","icudt54l.dat","samples","system","Tempdb","u54-msft_enus_0.txt","w32setupdb.cfg"
            break}
        'enterpriseserver' {
            $script:isServer=$True
            $script:InstallInfoProdType="Server"
            $script:installLogFile = join-path $env:TEMP "Zen_EnterpriseServer_PowerShell_Install.log"
            $script:filesToRemoveFromBin = "uwmkde.dll","w64repmkde.dll","psqlperfctrRE64.xml"
            $script:filesToRemoveFromZen = ""
            $script:filesToRemoveFromData = ""
            break}
        'reporting' {
            $script:isReporting=$True
            $script:InstallInfoProdType="Workgroup"
            $script:installLogFile = join-path $env:TEMP "Zen_ReportingEngine_PowerShell_Install.log"
            $script:filesToRemoveFromBin = "dbdefrag.exe","uwmkde.dll","ntmkde.dll","w64ntmkde.dll","psqlperfctrSE*.xml"
            $script:filesToRemoveFromZen = ""
            $script:filesToRemoveFromData = "defaultdb","Demodata","samples","system","w32setupdb.cfg"
            break}
    }
    # -- COM self-registering file lists.
    $script:ComFileList="acbtr732.ocx","acctr732.ocx","clientrb.dll","cobolschemaexecmsgrb.dll","dbcsipxy.dll","dto.dll","dto2.dll","legacylm.dll","licmgrrb.dll","mkc3.dll","mkderb.dll","pceurop.dll","pctlgrb.dll","poledb.dll","poledber.dll","poledbsy.dll","pscp932.dll","pscp936.dll","pscp949.dll","pscp950.dll","pscrb_en.dll","pscrb_ja.dll","psdom.dll","pseucjp.dll","pssax.dll","pstestct.dll","pstestsp.dll","pstsprb.dll","pvconnts_en.dll","pvconnts_ja.dll","pvinfodg_en.dll","pvinfodg_ja.dll","pvmsgrb.dll","srderb.dll","upiapirb.dll","w3cobolschemaexec100.dll","w3csi100.dll","w3csm100.dll","w3csp100.dll","w3dcm100.dll","w3exp010.dll","w3mpm100.dll","w64clientrb.dll","w64cobolschemaexec100.dll","w64cobolschemaexecmsgrb.dll","w64csi.dll","w64csm.dll","w64csp.dll","w64dbcsipxy.dll","w64dcm.dll","w64dto2.dll","w64exp010.dll","w64legacylm.dll","w64licmgrrb.dll","w64mkc3.dll","w64mkderb.dll","w64mpm.dll","w64pceurop.dll","w64pctlgrb.dll","w64pscp932.dll","w64pscp936.dll","w64pscp949.dll","w64pscp950.dll","w64psdom.dll","w64pseucjp.dll","w64pssax.dll","w64pvmsgrb.dll","w64srderb.dll","w64upiapirb.dll"
    # -- Obsolete file lists.
    $script:ObsoleteFileList="w3clnres.dll","w64clnres.dll","w3mgrres.dll","w64mgrres.dll","w3clnrsx.dll","w64clnrsx.dll","w3mgrrsx.dll","w64mgrrsx.dll"
}
function BeginLog {
    ##########################
    # Create $installLogFile and write basic info to the file.
    ##########################
    "#############################" | Out-File -Append $installLogFile
    $msg.infoStart -f [DateTime]::Parse((Get-Date).ToString()) | Out-File -Append $installLogFile
    $msg.infoScriptName -f $scriptName | Out-File -Append $installLogFile
    $msg.infoScriptPath -f $ScriptPath | Out-File -Append $installLogFile
    $msg.infoUpdating -f $updating | Out-File -Append $installLogFile
    $msg.infoCompName -f $computerName | Out-File -Append $installLogFile
    $msg.infoCulture -f $culture.Name | Out-File -Append $installLogFile
    $msg.infoProcNum -f [System.Environment]::ProcessorCount | Out-File -Append $installLogFile
    $msg.infoWinArch -f $env:PROCESSOR_ARCHITECTURE | Out-File -Append $installLogFile
    $msg.infoWinPlat -f [System.Environment]::OSVersion.Platform | Out-File -Append $installLogFile
    $msg.infoWinVer -f [System.Environment]::OSVersion.Version | Out-File -Append $installLogFile
    $msg.infoIs64Bit -f $Is64Bit | Out-File -Append $installLogFile
    $msg.infoCLRVer -f [System.Environment]::Version | Out-File -Append $installLogFile
    $msg.infoUserName -f $env:USERNAME | Out-File -Append $installLogFile
    $msg.infoUserDomain -f $env:USERDOMAIN | Out-File -Append $installLogFile
    $msg.infoPSVer -f $psVersionTable.PSVersion | Out-File -Append $installLogFile
    $msg.infoPSCLRVer -f $psVersionTable.CLRVersion | Out-File -Append $installLogFile
    $msg.infoFWDomainPS -f $fwDomainProfileState | Out-File -Append $installLogFile
    $msg.infoFWPrivatePS -f $fwPrivateProfileState | Out-File -Append $installLogFile
    $msg.infoFWPublicPS -f $fwPublicProfileState | Out-File -Append $installLogFile
    "" | Tee-Object -FilePath $installLogFile -Append
    $msg.infoRunning -f $scriptName, $InstallProductType | Tee-Object -FilePath $installLogFile -Append
    "" | Tee-Object -FilePath $installLogFile -Append
}
function StopPsqlServices {
    ##########################
    # Stop any running service with a display name beginning with the string
    # identified by $CmnSvcDispName.
    ##########################
    $msg.doStopSvcs -f $CmnSvcDispName | Tee-Object -FilePath $installLogFile -Append
    $foundSvc=$false
    get-service -ErrorAction Ignore | where {$_.DisplayName -Like $CmnSvcDispName+"*"} |
        ForEach-Object -Process {
            if ($_.Status -eq "Running") {
                $foundSvc=$true
                $svcDisplayName=$_.DisplayName
                Stop-Service -DisplayName $svcDisplayName -Force -ErrorAction Ignore | Out-Null
                if ($?) {
                    $msg.passStopSvcs -f $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
                } else {
                    $msg.warnStopSvcs -f $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
                }
                # Wait 5 seconds.
                Start-Sleep -Milliseconds 5000
            }
        }
    if (!$foundSvc) {$msg.notFoundRunningSvcs -f $CmnSvcDispName | Tee-Object -FilePath $installLogFile -Append}
    "" | Tee-Object -FilePath $installLogFile -Append
}
function UnpackFiles {
    ##########################
    # Unpacks the ZIP files associated with this script to the appropriate default Windows
    # location and then calls UpdateACLs to set the ACL for the folder to allow full
    # access for everyone.
    ##########################
    $msg.doUnpack | Tee-Object -FilePath $installLogFile -Append
    ## - Always uncompress DataFiles.zip.
    if (Test-Path -Path "$scriptPath\DataFiles.zip") {
        Expand-Archive -PATH "$scriptPath\DataFiles.zip" -DestinationPath $fpCommonApplicationData -Force
        if ($?) {
            $msg.passUnpack -f "$scriptPath\DataFiles.zip", $fpCommonApplicationData | Tee-Object -FilePath $installLogFile -Append
            UpdateACLs $ZenDataDir $defaultAclArgs
        } else {
            $msg.warnUnpack -f "$scriptPath\DataFiles.zip", $fpCommonApplicationData | Tee-Object -FilePath $installLogFile -Append
        }
    }
    ## - Program Files depends on bitness and architecture.
    if ($env:PROCESSOR_ARCHITECTURE -eq "ARM") {
        ## -- Windows on ARM is currently 32-bit only.
        if ($Is64Bit) { throw $msg.errNoArm64; exit -1 }
        $extractTo=$fpProgramFiles32
        $PgmFilesZip= join-path $scriptPath "ProgramFilesArm32.zip"
        $PgmFilesJaZip= join-path $scriptPath "ProgramFilesArm32_jajp.zip"
    } else {
        if ($Is64Bit) {
            ## -- 64-bit Windows.
            $extractTo=$fpProgramFiles64
            $PgmFilesZip= join-path $scriptPath "ProgramFilesX64.zip"
            $PgmFilesJaZip= join-path $scriptPath "ProgramFilesX64_jajp.zip"
        } else {
            ## -- 32-bit Windows.
            $extractTo=$fpProgramFiles32
            $PgmFilesZip= join-path $scriptPath "ProgramFilesX86.zip"
            $PgmFilesJaZip= join-path $scriptPath "ProgramFilesX86_jajp.zip"
        }
    }
    if (Test-Path -Path "$PgmFilesZip") {
        Expand-Archive -Path "$PgmFilesZip" -DestinationPath $extractTo -Force
        if ($?) {
            $msg.passUnpack -f "$PgmFilesZip", $extractTo | Tee-Object -FilePath $installLogFile -Append
            UpdateACLs "$extractTo\Actian" $defaultAclArgs
        } else {
            $msg.warnUnpack -f "$PgmFilesZip", $extractTo | Tee-Object -FilePath $installLogFile -Append
        }
    }
    if ($IsJajp) {
        if (Test-Path -Path "$PgmFilesJaZip") {
            Expand-Archive -Path "$PgmFilesJaZip" -DestinationPath $extractTo -Force
            if ($?) {
                $msg.passUnpack -f "$PgmFilesJaZip", $extractTo | Tee-Object -FilePath $installLogFile -Append
            } else {
                $msg.warnUnpack -f "$PgmFilesJaZip", $extractTo | Tee-Object -FilePath $installLogFile -Append
            }
        }
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function RemoveFilesByProductType {
    ##########################
    # For each product type, the list of files identified by the file lists below
    # are removed from the folders identified below:
    #    $filesToRemoveFromBin files are removed from the Bin folder.
    #    $filesToRemoveFromZen files are removed from the Zen folder (bin parent).
    #    $filesToRemoveFromData files are removed from the Data folder.
    ##########################
    $msg.doDelFiles -f $InstallProductType | Out-File -Append $installLogFile
    if ($Is64Bit) {$binDir=$ZenInstDir64+"bin"} else {$binDir=$ZenInstDir32+"bin"}
    if ($filesToRemoveFromBin -ne "") {
        $filesToRemoveFromBin | Foreach-Object -Process {
            if (Test-Path "$binDir\$_") {
                Try {
                    Remove-Item -Path "$binDir\$_" -Force
                    if ($?) { $msg.passDelFiles -f $InstallProductType, "$binDir\$_" | Out-File -Append $installLogFile }
                }
                Catch { $msg.warnDelFiles -f $InstallProductType, "$binDir\$_" | Out-File -Append $installLogFile }
            }
        }
    }
    if ($Is64Bit) {$zenDir=$ZenInstDir64} else {$zenDir=$ZenInstDir32}
    if ($filesToRemoveFromZen -ne "") {
        $filesToRemoveFromZen | Foreach-Object -Process {
            if (Test-Path "$zenDir\$_") {
                Try {
                    Remove-Item -Path "$zenDir\$_" -Force
                    if ($?) { $msg.passDelFiles -f $InstallProductType, "$zenDir\$_" | Out-File -Append $installLogFile }
                }
                Catch { $msg.warnDelFiles -f $InstallProductType, "$zenDir\$_" | Out-File -Append $installLogFile }
            }
        }
    }
    if ($filesToRemoveFromData -ne "") {
        $filesToRemoveFromData | Foreach-Object -Process {
            if (Test-Path $ZenDataDir$_) {
                Try {
                    Remove-Item -Path $ZenDataDir$_ -Recurse -Force
                    if ($?) { $msg.passDelFiles -f $InstallProductType, "$ZenDataDir$_" | Out-File -Append $installLogFile }
                }
                Catch { $msg.warnDelFiles -f $InstallProductType, "$ZenDataDir$_" | Out-File -Append $installLogFile }
            }
        }
    }
    "" | Out-File -Append $installLogFile
}
function UpdateACLs {
    ##########################
    # Updates the ACL for the specified file, folder, or registry key using the Get-Acl and
    # Set-Acl cmdlets and the .Net constructors for FileSystemAccessRule or RegistryAccessRule
    # to update the current ACLs to the specified ACLs according to the input parameters:
    # Assumes arguments are specified in the following order:
    #   $args[0]: Path to file, folder or registry key.
    #   $args[1]: ACL settings to be passed to the FileSystemAccessRule or RegistryAccessRule
    #             .net constructors.
    ##########################
    $ruleType="FileSystemAccessRule"
    if ($args.Count -ne 2) {
        $msg.warnUpdateACLs1 | Tee-Object -FilePath $installLogFile -Append
        return
    } else {
        $pathToUpdate=[string]$args[0]
        $aclArgs=$args[1]
    }
    $msg.doUpdateACLs -f [string]$aclArgs, $pathToUpdate | Tee-Object -FilePath $installLogFile -Append
    if (Test-Path $args[0]) {
        if ($pathToUpdate.IndexOf(":") -ge 4) {$ruleType="RegistryAccessRule"}
        try {
            $acl=Get-Acl $pathToUpdate
            $accessRule = New-Object System.Security.AccessControl.$ruleType $aclArgs
            $acl.SetAccessRule($accessRule)
            $acl | Set-Acl $pathToUpdate
            if ($?) { $msg.passUpdateACLs -f [string]$aclArgs, $pathToUpdate | Tee-Object -FilePath $installLogFile -Append }
        } catch {
            $msg.warnUpdateACLs3 -f $pathToUpdate, $aclArgs | Tee-Object -FilePath $installLogFile -Append
        }
    } else {
        $msg.warnUpdateACLs2 -f $args[0], $args[1] | Tee-Object -FilePath $installLogFile -Append
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function ReadFromConfig {
    ##########################
    # Reads the content of $ZenInstallCfg (same dir as script) and defines a global variable
    # for each specified name=value pair. If the variable is already defined it is
    # updated.
    ##########################
    $msg.doReadCfgFile -f $ZenInstallCfg | Tee-Object -FilePath $installLogFile -Append
    if (Test-Path "$scriptPath\$ZenInstallCfg") {
        # Set script-level properties.
        $props = Get-Content "$scriptPath\$ZenInstallCfg"
        For ($i=0; $i -lt $props.Count; $i++) {
            if ( $props[$i].StartsWith("#") ) {
                # Skip comment line.
            } else {
                if ( $props[$i].Contains("=") ) {
                    $items=$props[$i] -split "="
                    # Check if variable already set
                    Get-Variable $items[0] -ErrorAction Ignore | Out-Null
                    if ($?) {
                        $msg.addVarFromCfg -f $items[0], $items[1] | Out-File -Append $installLogFile
                        Set-Variable $items[0] -Value $items[1] -Scope Script #-errorAction Ignore
                    } else {
                        $msg.updateVarFromCfg -f $items[0], $items[1] | Out-File -Append $installLogFile
                        New-Variable $items[0] -Value $items[1] -Scope Script #-errorAction Ignore
                    }
                }
            }
        }
    } else {
        throw $msg.errNoCfgFile -f "$scriptPath\$ZenInstallCfg" | Tee-Object -FilePath $installLogFile -Append
        exit -1
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function UnloadPerfMonCounterSets {
    ##########################
    # Wrapper for Unlodctr.exe, called to unload the Zen counter sets for
    # Windows Performance Monitor.  This is only run during Enterprise Server
    # or Reporting Engine uninstalls and only if unlodctr.exe is found.
    #
    # Executes:
    #   "$env:windir\system32\unlodctr.exe" /m:{path}\{XML config file}
    ##########################
    # Skip if Client install.
    if ( $isClient) {
        # -- Windows Performance Monitor counter sets are not installed for WGE or Client.
        return
    }
    # Skip if system doesn't support it (no unlodctr.exe).
    $unloader="$env:windir\system32\unlodctr.exe"
    if (!(Test-Path $unloader)) {return}

    # Continue (unlodctr.exe found and SE, Vx, or Rpt engine).
    $msg.doUpdatePerfMonCtr | Tee-Object -FilePath $installLogFile -Append
    ## Set the configuration file.
    if ($isServer) {
        if ($Is64Bit) {
            $perfctrcfg=$ZenInstDir64 + "bin\psqlperfctrSE64.xml"
        } else {
            $perfctrcfg=$ZenInstDir32 + "bin\psqlperfctrSE32.xml"
        }
    } else {
        if ($isReporting) { $perfctrcfg=$ZenInstDir64 + "bin\psqlperfctrRE64.xml" }
    }
    if (Test-Path $perfctrcfg) {
        Try {
            $result=& UnlodCtr.exe /m:"$perfctrcfg" | Out-String -Stream
            if ($LASTEXITCODE -eq 0) {
                $msg.passUpdatePerfMonCtr | Tee-Object -FilePath $installLogFile -Append
            } else {
                $msg.warnUpdatePerfMonCtr -f $perfctrcfg, [string]$LASTEXITCODE, [string]$result | Tee-Object -FilePath $installLogFile -Append
            }
        }
        Catch { $msg.warnUpdatePerfMonCtr -f $perfctrcfg, [string]$LASTEXITCODE, [string]$result | Tee-Object -FilePath $installLogFile -Append }
    } else {
        $msg.warnNotFoundPerfMonCtrCfg | Tee-Object -FilePath $installLogFile -Append
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function RemoveObsoleteFilesOnUpgrade {
    ##########################
    # Invoked during an upgrade only. Removes any of the obsolete files identified by $ObsoleteFileList 
    ##########################
    $msg.doRemoveObsoleteFiles | Tee-Object -FilePath $installLogFile -Append
    if ($Is64Bit) {$filePath=$ZenInstDir64} else {$filePath=$ZenInstDir32}
    foreach ($fileToRemove in $ObsoleteFileList) {
        $fullFilePath=$filePath+"bin\"+$fileToRemove
        if (Test-Path $fullFilePath) {
            Remove-Item $fullFilePath -Force
            if ($?) {
                $msg.passRemoveObsoleteFiles -f $fullFilePath | Tee-Object -FilePath $installLogFile -Append
            } else {
                $msg.warnRemoveObsoleteFiles -f $fullFilePath | Tee-Object -FilePath $installLogFile -Append
            }
        }
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function UpdatePath {
    ##########################
    # Checks the PATH environment variable to determine if it already contains the
    # Zen bin folder path.  If not, the Zen bin folder path is added to the
    # beginning of the environment variable.
    #
    # Note: To keep consistent with the folder paths created by the MSI install, all
    #       folder paths added to the %PATH% below end with a backslash '\'.  If this
    #       is ever changed, the associated function in the uninstall script will
    #       also need to be modified.
    ##########################
    $msg.doUpdatePath | Tee-Object -FilePath $installLogFile -Append
    if ($Is64Bit) {$addToPath=$ZenInstDir64 + "bin"} else {$addToPath=$ZenInstDir32 + "bin"}
    # -- Update local machine PATH.
    $regPathParent="HKLM:System\CurrentControlSet\Control\Session Manager\Environment"
    $currentMachinePath=Get-ItemPropertyValue -Path $regPathParent -Name "Path"
    if ($currentMachinePath -like "*$addToPath*") {
        $msg.skipUpdatePath -f "Machine", $addToPath | Tee-Object -FilePath $installLogFile -Append
    } else {
        #Prepend bin location to PATH.
        Set-ItemProperty -Path $regPathParent -Name "Path" -Value "$addToPath\;$currentMachinePath"
        if ($?) {
            $msg.passUpdatePath -f $addToPath, "machine" | Tee-Object -FilePath $installLogFile -Append
        } else {
            $msg.warnUpdatePath -f $addToPath, "machine" | Tee-Object -FilePath $installLogFile -Append
        }
    }
    # -- Update current PowerShell session PATH.
    $currentSessionPath=[System.Environment]::GetEnvironmentVariable("Path")
    if ($currentSessionPath -like "*$addToPath*") {
        $msg.skipUpdatePath -f "Current session", $addToPath | Tee-Object -FilePath $installLogFile -Append
    } else {
        #Prepend bin location to PATH.
        [Environment]::SetEnvironmentVariable("Path", $addToPath +"\;" + $currentSessionPath)
        if ($?) {
            $msg.passUpdatePath -f $addToPath, "current session" | Tee-Object -FilePath $installLogFile -Append
        } else {
            $msg.warnUpdatePath -f $addToPath, "current session" | Tee-Object -FilePath $installLogFile -Append
        }
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function InitPsqlRegistry {
    ##########################
    # Create InstallInfo registry key and requisite registry values based on product type
    # specified on command line.  Note that all registry keys and values not explicitly
    # set in this procedure are created with their default values Zen is started for the
    # first time. If dbnames.cfg exists in one of the known previous data folder locations
    # listed below, the "DBNamesDirectory" registry value under "HKLM\SOFTWARE\Actian\Zen\Database Names"
    # will be created to point to the existing dbnames.cfg.
    # 
    # Known Previous Data Folder Locations:
    #   $PsqlDataDir
    #   $ZenDataDir
    #
    # Variables set:
    #   [object] $global:ZenInfo = see .OUTPUT in script header for a list of
    ##########################
    $msg.doPsqlReg  | Tee-Object -FilePath $installLogFile -Append
    ## Root key.
    if (Test-Path $registryPath) {
        $msg.skipCreateRootKey -f $registryPath | Tee-Object -FilePath $installLogFile -Append
    } else {
        $msg.addRootKey -f $registryPath | Tee-Object -FilePath $installLogFile -Append
        New-Item -Path $registryPath -Force | Out-Null
        UpdateACLs $registryPath $defaultAclArgs
    }
    if (Test-Path $registryPath"\InstallInfo\"$InstallInfoProdType"") {
        $msg.skipInstallInfoKey -f $registryPath | Tee-Object -FilePath $installLogFile -Append
    } else {
        $msg.addInstallInfoKey -f $registryPath | Tee-Object -FilePath $installLogFile -Append
        New-Item -Path $registryPath"\InstallInfo\"$InstallInfoProdType"" -Force | Out-Null
    }
    $msg.addInstallInfoValues -f $installLogFile | Tee-Object -FilePath $installLogFile -Append
    New-ItemProperty -Path $registryPath"\InstallInfo\"$InstallInfoProdType"" -Name "InstallData" -Value $ZenDataDir -PropertyType String -Force | Out-Null
    New-ItemProperty -Path $registryPath"\InstallInfo\"$InstallInfoProdType"" -Name "InstallDir" -Value $ZenInstDir64 -PropertyType String -Force | Out-Null
    New-ItemProperty -Path $registryPath"\InstallInfo\"$InstallInfoProdType"" -Name "InstallDir64" -Value $ZenInstDir64 -PropertyType String -Force | Out-Null
    if ($isServer) {New-ItemProperty -Path $registryPath"\InstallInfo\"$InstallInfoProdType"" -Name "IsVx" -Value "0" -PropertyType DWORD -Force | Out-Null}
    if ($isReporting) {New-ItemProperty -Path $registryPath"\InstallInfo\"$InstallInfoProdType"" -Name "IsBtrieve" -Value "0" -PropertyType DWORD -Force | Out-Null}
    if ($isReporting) {New-ItemProperty -Path $registryPath"\InstallInfo\"$InstallInfoProdType"" -Name "IsReport" -Value "1" -PropertyType DWORD -Force | Out-Null}
    if ($Is64Bit -and $isServer) {New-ItemProperty -Path $registryPath"\InstallInfo\"$InstallInfoProdType"" -Name "Server64" -Value "YES" -PropertyType String -Force | Out-Null}
    # -- Use default values if properties not read from configuration file.
    New-ItemProperty -Path $registryPath"\InstallInfo\"$InstallInfoProdType"" -Name "ProductVersion" -Value $ProductVer -PropertyType String -Force | Out-Null
    New-ItemProperty -Path $registryPath"\InstallInfo\"$InstallInfoProdType"" -Name "VersionLevel" -Value $VersionLevel -PropertyType String -Force | Out-Null
    # Set properties for global object $ZenInfo to InstallInfo registry values.
    $global:ZenInfo=Get-ItemProperty $registryPath"\InstallInfo\"$InstallInfoProdType""
    ## ELS Key
    if (Test-Path $registryPath"\ELS") {
        $msg.skipElsKey -f "$registryPath\ELS" | Tee-Object -FilePath $installLogFile -Append
    } else {
        $msg.addElsKey -f "$registryPath\ELS" | Tee-Object -FilePath $installLogFile -Append
        New-Item -Path $registryPath"\ELS" -Force | Out-Null
    }
    $msg.addElsValues -f $installLogFile | Tee-Object -FilePath $installLogFile -Append
    New-ItemProperty -Path $registryPath"\ELS" -Name "elsh" -Value $zenElsURL -PropertyType String -Force | Out-Null
    New-ItemProperty -Path $registryPath"\ELS" -Name "elshid" -Value $zenElsServerID -PropertyType DWORD -Force | Out-Null
    New-ItemProperty -Path $registryPath"\ELS" -Name "logfile" -Value (join-path $ZenDataDir "logs\els.log") -PropertyType String -Force | Out-Null
    
    ## DBNamesDirectory registry value if previous dbnames.cfg exists.
    if (Test-Path (join-path $PsqlDataDir "DBNAMES.CFG")) {
        $msg.addDBNamesDirectory -f $PsqlDataDir | Tee-Object -FilePath $installLogFile -Append
        if (Test-Path (join-path $registryPath "Database Names")) {
            New-ItemProperty -Path (join-path $registryPath "Database Names") -Name "DBNamesDirectory" -Value $PsqlDataDir -PropertyType String -Force | Out-Null
        } else {
            New-Item -Path (join-path $registryPath "Database Names") -Force | Out-Null
            New-ItemProperty -Path (join-path $registryPath "Database Names") -Name "DBNamesDirectory" -Value $PsqlDataDir -PropertyType String -Force | Out-Null
        }
    }
    if (Test-Path (join-path $ZenDataDir "DBNAMES.CFG")) {
        $msg.addDBNamesDirectory -f $ZenDataDir | Tee-Object -FilePath $installLogFile -Append
        if (Test-Path (join-path $registryPath "Database Names")) {
            New-ItemProperty -Path (join-path $registryPath "Database Names") -Name "DBNamesDirectory" -Value $ZenDataDir -PropertyType String -Force | Out-Null
            "found Database Names key"
        } else {
            "didn't find Database Names key"
            New-Item -Path (join-path $registryPath "Database Names") -Force | Out-Null
            New-ItemProperty -Path (join-path $registryPath "Database Names") -Name "DBNamesDirectory" -Value $ZenDataDir -PropertyType String -Force | Out-Null
        }
    }
    # Output current product registry hive to logfile.
    "-----------------------------------------------------------------------------------------" | Out-File -Append $installLogFile
    Get-ChildItem $registryPath | Out-File -Append $installLogFile
    "-----------------------------------------------------------------------------------------" | Out-File -Append $installLogFile
    Get-ItemProperty "$registryPath\InstallInfo\$InstallInfoProdType" | Out-File -Append $installLogFile
    "-----------------------------------------------------------------------------------------" | Out-File -Append $installLogFile
    "" | Tee-Object -FilePath $installLogFile -Append
}
function RegisterComInterfaces {
    ##########################
    # Checks to see which Zen COM interface library files are installed and then registers
    # each file using the regsvr32.exe utility.
    ##########################
    $msg.doRegisterCom  | Tee-Object -FilePath $installLogFile -Append
    $nFound=0;
    $nRegistered=0;
    if ($Is64Bit) {$filePath=$ZenInstDir64} else {$filePath=$ZenInstDir32}
    foreach ($fileToRegister in $ComFileList) {
        $fullFilePath=$filePath+"bin\"+$fileToRegister
        if (Test-Path $fullFilePath) {
            $nFound++
            & regsvr32.exe -s "$fullFilePath"| Out-String -Stream
            switch ($LASTEXITCODE) {
                0       {$msg.passRegFile -f $fullFilePath | Out-File -Append $installLogFile; $nRegistered++;break}
                default {$msg.warnRegFile -f $fullFilePath | Tee-Object -FilePath $installLogFile -Append; break}
            }
        }
    }
    "Successfully self-registered $nRegistered out of $nFound files. Refer to $installLogFile for more information." | Tee-Object -FilePath $installLogFile -Append
    "" | Tee-Object -FilePath $installLogFile -Append
}
function CreateServiceIfNotExist {
    ##########################
    # Wrapper for New-Service commandlet that creates a new service only after
    # first checking to verify the specified binary exists and the service
    # doesn't already exist.
    # Assumes arguments are specified in the following order:
    #   $args[0]: Path to binary.
    #   $args[1]: Service name
    #   $args[2]: Service display name
    #   $args[3]: Service description
    #   $args[4]: service arguments (if any)
    ##########################
    $svcPath=[string]$args[0]+[string]$args[4]
    $svcName=[string]$args[1]
    $svcDisplayName=[string]$args[2]
    $svcDesc=[string]$args[3]
    if (Test-Path $args[0]) {
        if (get-Service -Name $svcName -errorAction Ignore) {
            $msg.skipAddSvc -f $svcName | Tee-Object -FilePath $installLogFile -Append
        } else {
            New-Service -BinaryPathName $svcPath -Name $svcName -DisplayName $svcDisplayName -Description $svcDesc  -StartupType Automatic  -ErrorAction Ignore | Out-Null
            if ($?) {
               $msg.passAddSvc -f $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
            } else {
                $msg.errAddSvc -f $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
                "" | Tee-Object -FilePath $installLogFile -Append
                exit -1
            }
        }
    } else {
        $msg.warnNoSvcFile -f [string]$args[0], $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
    }
}
function CreatePsqlServices {
    ##########################
    # Specifies the service parameters for each product type and then passes
    # them to CreateServiceIfNotExist to verify requires, create the service
    # if it doesn't already exist, and log transactions.
    ##########################
    if ( $isClient) {
        # -- No services are created for client installs.
        return
    }
    $msg.doAddSvc | Tee-Object -FilePath $installLogFile -Append
    if ($Is64Bit) {
        if ($isServer) {
            # 64-bit Enterprise Server
            $svcPath=$ZenInstDir64 + "bin\zenengnsvc.exe"
            $svcName="zenengine"
            $svcDisplayName="Actian Zen Enterprise Server"
            $svcDesc=$msg.svcDesc -f $svcDisplayName
            $svcArgs=""
            CreateServiceIfNotExist $svcPath $svcName $svcDisplayName $svcDesc $svcArgs
        }
        if ($isReporting) {
            # 64-bit Reporting Engine
            $svcPath=$ZenInstDir64 + "bin\zenengnsvc.exe"
            $svcName="zenengine"
            $svcDisplayName="Actian Zen Reporting Engine"
            $svcDesc=$msg.svcDesc -f $svcDisplayName
            $svcArgs=" -service -srde"
            CreateServiceIfNotExist $svcPath $svcName $svcDisplayName $svcDesc $svcArgs
        }
    } else {
        if ($isServer) {
            # 32-bit Enterprise Server
            $svcPath=$ZenInstDir32 + "bin\zenengnsvc32.exe"
            $svcName="zenengine"
            $svcDisplayName="Actian Zen Enterprise Server"
            $svcDesc=$msg.svcDesc -f $svcDisplayName
            $svcArgs=""
            CreateServiceIfNotExist $svcPath $svcName $svcDisplayName $svcDesc $svcArgs
        }
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function AddFirewallRuleIfNotExist {
    ##########################
    # Create the specified firewall rule only if the specified binary exists and
    # the specified rule name does not already exist.  The specified rule is
    # created but is NOT ENABLED by default.
    # Assumes arguments are specified in the following order:
    #   $args[0]: Path to binary
    #   $args[1]: rule name
    ##########################
    $rulePath=$args[0]
    $ruleName=$args[1]
    if (Test-Path $rulePath) {
        Try {
            # Use "-ErrorAction Stop" to trigger failure when Get-NetFirewallRule cmdlet fails.
            Get-NetFirewallRule -DisplayName $ruleName -ErrorAction Stop | Out-Null
            $msg.skipAddRule -f $ruleName | Tee-Object -FilePath $installLogFile -Append
        }
        Catch {
            # -- Rule doesn't exist, create it. Always disable the rule for all firewall profiles.
            Try {
                New-NetFirewallRule -DisplayName $ruleName -Direction Inbound -Program $rulePath -Action Allow -Profile Any -Enabled False | Out-File -Append $installLogFile
                if ($?) { $msg.passAddRule -f $ruleName | Tee-Object -FilePath $installLogFile -Append }
            }
            Catch {$msg.warnAddRuleFailed -f $ruleName | Tee-Object -FilePath $installLogFile -Append}
        }

    } else {$msg.warnNoRuleFile -f $rulePath, $ruleName | Tee-Object -FilePath $installLogFile -Append}
}
function AddPsqlFirewallRules {
    ##########################
    # If the NetSecurity module is present this function defines firewall rule
    # properties based on product type and then calls AddFirewallRuleIfNotExist
    # to create the specified rule(s) for the current product type.
    ##########################
    if ( $isClient) {
        # -- No firewall rules are created for client installs.
        return
    }
    $msg.doAddFWRules | Tee-Object -FilePath $installLogFile -Append
    $moduleNetSecurity=$false
    Get-Command -CommandType Function | ForEach-Object -Process {if ($_.Source -eq "NetSecurity") {$moduleNetSecurity=$true} }
    if ($moduleNetSecurity) {
        Import-Module -Name NetSecurity  # -- Firewall rule cmdlets.
    } else {
        # -- The NetSecurity Module is not installed, warn user and exit function.
        $msg.warnNoNetSecCreate | Tee-Object -FilePath $installLogFile -Append
        "" | Tee-Object -FilePath $installLogFile -Append
        return
    }
    if ($Is64Bit) {
      if ($isServer) {
          $fwRulePath=$ZenInstDir64 + "bin\zenengnsvc.exe"
          $fwRuleName="Actian Zen Server Engine"
          AddFirewallRuleIfNotExist $fwRulePath $fwRuleName
      }
      if ($isReporting) {
          $fwRulePath=$ZenInstDir64 + "bin\zenengnsvc.exe"
          $fwRuleName="Actian Zen Reporting Engine"
          AddFirewallRuleIfNotExist $fwRulePath $fwRuleName
      }
    } else {
      if ($isServer) {
          $fwRulePath=$ZenInstDir32 + "bin\zenengnsvc32.exe"
          $fwRuleName="Actian Zen Server Engine"
          AddFirewallRuleIfNotExist $fwRulePath $fwRuleName
      }
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function StartPsqlServices {
    ##########################
    # Start any service with a display name beginning with the string
    # identified by $CmnSvcDispName.
    ##########################
    $msg.doStartPsqlSvcs | Tee-Object -FilePath $installLogFile -Append
    $foundSvc=$false
    get-service | where {$_.DisplayName -Like $CmnSvcDispName+"*"} | Sort-Object -Descending |
        ForEach-Object -Process {
            $foundSvc=$true
            if ($_.Status -ne "Running") {
                $svcDisplayName=$_.DisplayName
                $svcName=$_.Name
                Start-Service -Name $svcName -ErrorAction Ignore| Out-Null
                if ($?) {
                    $msg.passStartSvcs -f $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
                } else {
                    $msg.errStartSvcs -f $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
                    "" | Tee-Object -FilePath $installLogFile -Append
                    exit -1
                }
                # Wait 5 seconds.
                Start-Sleep -Milliseconds 5000
            } else {
                $msg.skipStartSvcs -f $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
            }
        }
    if (!$foundSvc) {$msg.notFoundSvcs -f $CmnSvcDispName | Tee-Object -FilePath $installLogFile -Append}
    "" | Tee-Object -FilePath $installLogFile -Append
}
function CreatePsqlDatabases {
    ##########################
    # Runs the command line setupdbcli.exe utility to create the system databases, bulk load data
    # into them, and execute SQL statements to configure them as determined by the settings
    # specified in the configuraiton file being processed.
    #
    # PREREQ: The Zen Service(s) must be started prior to calling setupdbcli.exe.
    ##########################
    if ( $isReporting -or $isClient) {
        # -- No databases are created during install for client or reporting engine.
        $msg.skipCreateDbs | Tee-Object -FilePath $installLogFile -Append
        "" | Tee-Object -FilePath $installLogFile -Append
        return
    }

    # -- Continue (Enterprise Server or Workgroup only).
    $msg.doCreateDbs | Tee-Object -FilePath $installLogFile -Append

    # -- Verify setupdb configuration file exists.
    $setupdbLog=$ZenDataDir+"Logs\Zen_Setupdb.log"
    if (Test-Path $setupdbCfgInstall) {
        $setupdb_args="-log",$setupdbLog,"-dir",$ZenDataDir,$setupdbCfgInstall
    } else {
        $msg.noSetupDbCfgCreate -f $setupdbCfgInstall | Tee-Object -FilePath $installLogFile -Append
    }
    if (Test-Path $PsqlDataDir"DBNAMES.CFG") {
        $msg.foundDbnames -f $PsqlDataDir | Tee-Object -FilePath $installLogFile -Append
    } else {
        if (Test-Path $ZenDataDir"DBNAMES.CFG") {
            $msg.foundDbnames -f $ZenDataDir | Tee-Object -FilePath $installLogFile -Append
        } else {
            $msg.notFoundDbnames -f $ZenDataDir | Tee-Object -FilePath $installLogFile -Append
        }
    }

    # -- Execute setupdbcli.exe.
    if ($Is64Bit) {
        if ($isServer) {
            if (Test-Path $ZenInstDir64"bin\setupdbcli.exe") {
                # -- No return code from setupdbcli.exe so just log command line.
                $msg.runningSetupdb -f [string]$setupdb_args, $setupdbLog | Tee-Object -FilePath $installLogFile -Append
                Try {
                    Start-Process -FilePath $ZenInstDir64"bin\setupdbcli.exe" -ArgumentList $setupdb_args -WorkingDirectory $ZenInstDir64"bin" -NoNewWindow -Wait
                    if ($?) { $msg.passSetupdb -f $ZenInstDir64, $setupdbLog | Tee-Object -FilePath $installLogFile -Append }
                }
                Catch  { $msg.failSetupdb -f $ZenInstDir64 | Tee-Object -FilePath $installLogFile -Append }
            } else {
                $msg.noSetupDb -f $ZenInstDir64 | Tee-Object -FilePath $installLogFile -Append
            }
        }
    } else {
        if (Test-Path $ZenInstDir32"bin\setupdbcli.exe") {
            # -- No return code from setupdbcli.exe so just log command line.
            $msg.runningSetupdb -f [string]$setupdb_args, $setupdbLog | Tee-Object -FilePath $installLogFile -Append
            Try {
                Start-Process -FilePath $ZenInstDir32"bin\setupdbcli.exe" -ArgumentList $setupdb_args -WorkingDirectory $ZenInstDir32"bin" -NoNewWindow -Wait
                if ($?) { $msg.passSetupdb -f $ZenInstDir32, $setupdbLog | Tee-Object -FilePath $installLogFile -Append }
            }
            Catch  {  $msg.failSetupdb -f $ZenInstDir32 | Tee-Object -FilePath $installLogFile -Append }
        } else {
            $msg.noSetupDb -f $ZenInstDir32 | Tee-Object -FilePath $installLogFile -Append
        }
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function ActivatePsqlTempLicense {
    ##########################
    # Calls the appropriate command line license administrator utility to add the temp
    # license key defined in $ZenInstallCfg and stored in $PVSW_PSQL_LICENSE_KEY per
    # the previous call to the ReadFromConfig function.
    ##########################
    if ( $isReporting -or $isClient -or $updating) {
        # -- No license required for client or reporting engine or if updating.
        return
    }

    # -- Continue (Enterprise Server or Workgroup only).
    $msg.doAddTempLic | Tee-Object -FilePath $installLogFile -Append
    $result=$false
    if ($Is64Bit) {
        if (Test-Path $ZenInstDir64"bin\clilcadm.exe") {
            $result=& clilcadm.exe -a $PVSW_PSQL_LICENSE_KEY| Out-String -Stream
        } else {
            $msg.notFoundClilcadm -f $ZenInstDir64 | Tee-Object -FilePath $installLogFile -Append
        }
    } else {
        if (Test-Path $ZenInstDir32"bin\clilcadm.exe") {
            $result=& clilcadm.exe -a $PVSW_PSQL_LICENSE_KEY| Out-String -Stream
        } else {
            $msg.notFoundClilcadm -f $ZenInstDir32 | Tee-Object -FilePath $installLogFile -Append
        }
    }

    # -- Interpret results.
    if ($result) {
        switch ($lastExitCode) {
            0       {$msg.passAddTempLic -f $PVSW_PSQL_LICENSE_KEY | Tee-Object -FilePath $installLogFile -Append; break}
            default {$msg.warnAddTempLic -f $result[4], $result[5], $result[6] | Tee-Object -FilePath $installLogFile -Append; break}
        }
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function InvokeECollect {
    ##########################
    # Wrapper function for making DTI call pvCollectLocal() to collect the
    # Zen product information during install or update.  This call requires
    # the appropriate ELS library is installed so we skip this step if the
    # library is not found.
    #
    # API Invoked
    #   32-bit systems: clientlm.dll (requires w3els.dll)
    #   64-bit systems: w64clientlm.dll (requires w64els.dll)
    #     install - pvCollectLocal(COLLECT_ON_INSTALL)
    #     update  - pvCollectLocal(COLLECT_ON_PATCH)
    ##########################

    # Skip eCollect if the appropriate ELS library is not present or if this is
    # a client or Reporting Engine install (pvCollectLocal only supported during engine installs).
    if ($Is64Bit) { $elsLib=$ZenInstDir64 + "bin\w64els.dll"} else {$elsLib=$ZenInstDir32 + "bin\w3els.dll"}
    if (!(Test-Path $elsLib) -or $isClient -or $isReporting ) {
        $msg.skipECollect | Tee-Object -FilePath $installLogFile -Append
        "" | Tee-Object -FilePath $installLogFile -Append
        return
    }

    # Not sure why but the ELS libraries are requiring the Zen services to be restarted
    # on install after the system databases are created during CreatePsqlDatabases().
    StopPsqlServices
    StartPsqlServices

    $msg.doECollect | Tee-Object -FilePath $installLogFile -Append
    if ($updating) {[int]$eventId=120} else {[int]$eventId=100}
    if ($Is64Bit) {$dllName="w64clientlm.dll"} else {$dllName="clientlm.dll"}

    # Create the eCollect ScriptBlock. We need to invoke the ScriptBlock as a
    # background task because PowerShell has no way of unloading a DLL after the
    # Add-Type cmdlet has been used, resulting in w64/clientlm.dll and all
    # dependent libraries being locked.
    # Note the two lines in the 'here string' that identify the ELS library
    # signature can't have any leading spaces.
    $apiScriptBlock= {
        param ([string]$dllFileName="clientlm.dll", [int]$eCollectEventId=100)
        $ApiSignature = @"
[DllImport("$dllFileName")]
public static extern int pvCollectLocal(int nEventId);
"@
        Try {
            $pvCollectLocal= Add-Type -MemberDefinition $ApiSignature -Name "Win32pvCollectLocal" -Namespace Win32Functions -PassThru
            $pvCollectLocal::pvCollectLocal($eCollectEventId)
        }
        Catch {[string]$error[0]}
    }

    # Execute the scriptblock with the appropriate library name and event ID.
    Try {
        Start-Job -Name eCollect -ScriptBlock $apiScriptBlock -args $dllName,$eventId | Out-Null
        Wait-Job -Name eCollect | Out-Null
        $result=Receive-Job -Name eCollect
        if ($result.GetType().ToString() -eq "System.Int32") {
            # ScriptBlock returned int.
            if ($result -eq 0) {
                $msg.passECollect | Tee-Object -FilePath $installLogFile -Append
            } else {
                $msg.warnECollect -f "pvCollectLocal call",$result | Tee-Object -FilePath $installLogFile -Append
            }
        } else {
           # ScriptBlock returned non-int.
           $msg.warnECollect -f "eCollect ScriptBlock",$result | Tee-Object -FilePath $installLogFile -Append
        }
    }
    Catch {$msg.warnECollect -f "eCollect ScriptBlock",$result | Tee-Object -FilePath $installLogFile -Append}
    "" | Tee-Object -FilePath $installLogFile -Append
}
function UpdateRoyaltyRecord {
    ##########################
    # Wrapper function for making DTI call PvValidateLicenses() to update the
    # royalty record (also creates the 'Product Data' registry key).
    #
    # API Invoked
    #   32-bit systems: w3dbav90.dll::PvValidateLicenses(0xFFFFFFFF)
    #   64-bit systems: w64dba.dll::PvValidateLicenses(0xFFFFFFFF)
    ##########################

    # Skip if this is a client or reporting engine install (PvValidateLicenses only supported during engine installs).
    if ( $isClient  -or $isReporting) {return}

    $msg.doUpdateRoyaltyRec | Tee-Object -FilePath $installLogFile -Append
    if ($Is64Bit) {$dllName="w64dba.dll"} else {$dllName="w3dbav90.dll"}

    # Create the ScriptBlock to update the royalty record. We need to invoke
    # the ScriptBlock as a background task because PowerShell has no way of
    # unloading a DLL after the Add-Type cmdlet has been used, resulting in
    # library file and all dependent libraries being locked.
    # Note the two lines in the 'here string' that identify the DTI library
    # signature can't have any leading spaces.
    $apiScriptBlock= {
        param ([string]$dllFileName="w3dbav90.dll")
        $ApiSignature = @"
[DllImport("$dllFileName")]
public static extern int PvValidateLicenses(int hConnection);
"@
        Try {
            [long]$hConnection=0xFFFFFFFF
            $PvValidateLicenses= Add-Type -MemberDefinition $ApiSignature -Name "Win32pvCollectLocal" -Namespace Win32Functions -PassThru
            $PvValidateLicenses::PvValidateLicenses($hConnection)
        }
        Catch {[string]$error[0]}
    }

    # Execute the scriptblock with the appropriate library name and event ID.
    Try {
        Start-Job -Name eCollect -ScriptBlock $apiScriptBlock -args $dllName | Out-Null
        Wait-Job -Name eCollect | Out-Null
        $result=Receive-Job -Name eCollect
        if ($result.GetType().ToString() -eq "System.Int32") {
            # ScriptBlock returned int.
            if ($result -eq 0) {
                $msg.passUpdateRoyaltyRec | Tee-Object -FilePath $installLogFile -Append
            } else {
                $msg.warnUpdateRoyaltyRec -f "PvValidateLicenses call",$result | Tee-Object -FilePath $installLogFile -Append
            }
        } else {
           # ScriptBlock returned non-int.
           $msg.warnUpdateRoyaltyRec -f "Update royalty record ScriptBlock",$result | Tee-Object -FilePath $installLogFile -Append
        }
    }
    Catch {$msg.warnUpdateRoyaltyRec -f "eCollect ScriptBlock",$result | Tee-Object -FilePath $installLogFile -Append}
    "" | Tee-Object -FilePath $installLogFile -Append
}
function LoadPerfMonCounterSets {
    ##########################
    # Wrapper for Lodctr.exe, called to load the Zen counter sets for
    # Windows Performance Monitor.  This is only run during Enterprise Server
    # or Reporting Engine installs and only if lodctr.exe is found.
    #
    # Executes:
    #   "$env:windir\system32\lodctr.exe" {path}\{XML config file}
    ##########################
    # Skip if Client install.
    if ( $isClient) {
        # -- Windows Performance Monitor counter sets are not installed for Client.
        return
    }
    # Skip if system doesn't support it (no lodctr.exe).
    $loader="$env:windir\system32\lodctr.exe"
    if (!(Test-Path $loader)) {return}

    # Continue (lodctr.exe found and SE, Vx, or Rpt engine).
    $msg.doLoadPerfMonCtr | Tee-Object -FilePath $installLogFile -Append
    ## Set the configuration file.
    if ($isServer) {
        if ($Is64Bit) {
            $perfctrcfg=$ZenInstDir64 + "bin\psqlperfctrSE64.xml"
        } else {
            $perfctrcfg=$ZenInstDir32 + "bin\psqlperfctrSE32.xml"
        }
    } else {
        if ($isReporting) { $perfctrcfg=$ZenInstDir64 + "bin\psqlperfctrRE64.xml" }
    }
    if (Test-Path $perfctrcfg) {
        Try {
            $result=& LodCtr.exe /m:"$perfctrcfg" | Out-String -Stream
            if ($LASTEXITCODE -eq 0) {
                $msg.passLoadPerfMonCtr | Tee-Object -FilePath $installLogFile -Append
            } else {
                $msg.warnLoadPerfMonCtr -f $perfctrcfg, [string]$LASTEXITCODE, [string]$result | Tee-Object -FilePath $installLogFile -Append
            }
        }
        Catch { $msg.warnLoadPerfMonCtr -f $perfctrcfg, [string]$LASTEXITCODE, [string]$result | Tee-Object -FilePath $installLogFile -Append }
    } else {
        $msg.warnNotFoundPerfMonCtrCfg | Tee-Object -FilePath $installLogFile -Append
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function EndLog {
    ##########################
    # Close $installLogFile with date/time script completed.
    ##########################
    $msg.infoEnd -f [DateTime]::Parse((Get-Date).ToString()) | Out-File -Append $installLogFile
    ##########################
    "#############################" | Tee-Object -FilePath $installLogFile -Append
    "" | Out-File -Append $installLogFile
    $msg.completeInstallMsg -f $installLogFile
    "" | Tee-Object -FilePath $installLogFile -Append
}

Try {. Main}
Catch {throw $msg.errUnknown; exit -1}

# SIG # Begin signature block
# MIIcMwYJKoZIhvcNAQcCoIIcJDCCHCACAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUg7+45kFbFYa4rMCcd2pOazB7
# n6igghcYMIID7jCCA1egAwIBAgIQfpPr+3zGTlnqS5p31Ab8OzANBgkqhkiG9w0B
# AQUFADCBizELMAkGA1UEBhMCWkExFTATBgNVBAgTDFdlc3Rlcm4gQ2FwZTEUMBIG
# A1UEBxMLRHVyYmFudmlsbGUxDzANBgNVBAoTBlRoYXd0ZTEdMBsGA1UECxMUVGhh
# d3RlIENlcnRpZmljYXRpb24xHzAdBgNVBAMTFlRoYXd0ZSBUaW1lc3RhbXBpbmcg
# Q0EwHhcNMTIxMjIxMDAwMDAwWhcNMjAxMjMwMjM1OTU5WjBeMQswCQYDVQQGEwJV
# UzEdMBsGA1UEChMUU3ltYW50ZWMgQ29ycG9yYXRpb24xMDAuBgNVBAMTJ1N5bWFu
# dGVjIFRpbWUgU3RhbXBpbmcgU2VydmljZXMgQ0EgLSBHMjCCASIwDQYJKoZIhvcN
# AQEBBQADggEPADCCAQoCggEBALGss0lUS5ccEgrYJXmRIlcqb9y4JsRDc2vCvy5Q
# WvsUwnaOQwElQ7Sh4kX06Ld7w3TMIte0lAAC903tv7S3RCRrzV9FO9FEzkMScxeC
# i2m0K8uZHqxyGyZNcR+xMd37UWECU6aq9UksBXhFpS+JzueZ5/6M4lc/PcaS3Er4
# ezPkeQr78HWIQZz/xQNRmarXbJ+TaYdlKYOFwmAUxMjJOxTawIHwHw103pIiq8r3
# +3R8J+b3Sht/p8OeLa6K6qbmqicWfWH3mHERvOJQoUvlXfrlDqcsn6plINPYlujI
# fKVOSET/GeJEB5IL12iEgF1qeGRFzWBGflTBE3zFefHJwXECAwEAAaOB+jCB9zAd
# BgNVHQ4EFgQUX5r1blzMzHSa1N197z/b7EyALt0wMgYIKwYBBQUHAQEEJjAkMCIG
# CCsGAQUFBzABhhZodHRwOi8vb2NzcC50aGF3dGUuY29tMBIGA1UdEwEB/wQIMAYB
# Af8CAQAwPwYDVR0fBDgwNjA0oDKgMIYuaHR0cDovL2NybC50aGF3dGUuY29tL1Ro
# YXd0ZVRpbWVzdGFtcGluZ0NBLmNybDATBgNVHSUEDDAKBggrBgEFBQcDCDAOBgNV
# HQ8BAf8EBAMCAQYwKAYDVR0RBCEwH6QdMBsxGTAXBgNVBAMTEFRpbWVTdGFtcC0y
# MDQ4LTEwDQYJKoZIhvcNAQEFBQADgYEAAwmbj3nvf1kwqu9otfrjCR27T4IGXTdf
# plKfFo3qHJIJRG71betYfDDo+WmNI3MLEm9Hqa45EfgqsZuwGsOO61mWAK3ODE2y
# 0DGmCFwqevzieh1XTKhlGOl5QGIllm7HxzdqgyEIjkHq3dlXPx13SYcqFgZepjhq
# IhKjURmDfrYwggR9MIIDZaADAgECAgMb5xUwDQYJKoZIhvcNAQELBQAwYzELMAkG
# A1UEBhMCVVMxITAfBgNVBAoTGFRoZSBHbyBEYWRkeSBHcm91cCwgSW5jLjExMC8G
# A1UECxMoR28gRGFkZHkgQ2xhc3MgMiBDZXJ0aWZpY2F0aW9uIEF1dGhvcml0eTAe
# Fw0xNDAxMDEwNzAwMDBaFw0zMTA1MzAwNzAwMDBaMIGDMQswCQYDVQQGEwJVUzEQ
# MA4GA1UECBMHQXJpem9uYTETMBEGA1UEBxMKU2NvdHRzZGFsZTEaMBgGA1UEChMR
# R29EYWRkeS5jb20sIEluYy4xMTAvBgNVBAMTKEdvIERhZGR5IFJvb3QgQ2VydGlm
# aWNhdGUgQXV0aG9yaXR5IC0gRzIwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEK
# AoIBAQC/cWII8fpZNPcbyRij94BJWOkigxOmxSBDATuE8eaFSZ8n6vaEG06gtNtw
# mMcyAbEFPgdO7vT6Ty9ZMCLnqxlWa+KAB/zzFnWAOVF75fk1tnROqY2CE+S2P6kD
# g/qivooVan/eC8O2GRQFyurDqASUO0Z8Mg3zAGYiyI1pbTaMERi307IcYLQ4+gKM
# ztPdRgfeCj7rXXzIfPuwK1OkkmJpUSUFYRpEgYwsqUOWI9+sOoGaDinFHKnpXR62
# np4wCjnO8YiA+0tdzDLshWJDJTQCVicBkbQ7cCo/brHonIgBfZ/U+dtTbWCdvyzn
# WKu4X0b8zsQbAzwJ60kxXGlGs+BHAgMBAAGjggEXMIIBEzAPBgNVHRMBAf8EBTAD
# AQH/MA4GA1UdDwEB/wQEAwIBBjAdBgNVHQ4EFgQUOpqFBxBnKLbv9r0FQW4gwZTa
# D94wHwYDVR0jBBgwFoAU0sSw0pHUTBFxs2HLPaH+3ahq1OMwNAYIKwYBBQUHAQEE
# KDAmMCQGCCsGAQUFBzABhhhodHRwOi8vb2NzcC5nb2RhZGR5LmNvbS8wMgYDVR0f
# BCswKTAnoCWgI4YhaHR0cDovL2NybC5nb2RhZGR5LmNvbS9nZHJvb3QuY3JsMEYG
# A1UdIAQ/MD0wOwYEVR0gADAzMDEGCCsGAQUFBwIBFiVodHRwczovL2NlcnRzLmdv
# ZGFkZHkuY29tL3JlcG9zaXRvcnkvMA0GCSqGSIb3DQEBCwUAA4IBAQBZC1O9koYR
# pyR77Vsxzx0fbHDFuG6+Trv2vpdQ4TB/uihcYpTC434z9/tCdoXblRyMIlh1CQyI
# ZWc5ChYJxaA4l6TFI5M/tBimAQZEkeOnaSe0WiV/Orcyzd2E/yo4KTOk3Weyhf6h
# iCAcUInI3Cr2QgM3TOaI39WvJPKxw9/MtezgmV63SVQgPJQYDMccUhhJpG3hs1gL
# ydjs2a4cMo4ocA3i/qYXnoQPvVdws1rpH6CGU7vvfP9pC+BIw7eTC8gKVMSsXRRn
# N2zKpS8xCDeqbm+MvJviV10kga+Xl5yErWysN0xm82GRESDkvjCfeqQpCbDhNF9k
# dxhAUd+MMKavMIIEozCCA4ugAwIBAgIQDs/0OMj+vzVuBNhqmBsaUDANBgkqhkiG
# 9w0BAQUFADBeMQswCQYDVQQGEwJVUzEdMBsGA1UEChMUU3ltYW50ZWMgQ29ycG9y
# YXRpb24xMDAuBgNVBAMTJ1N5bWFudGVjIFRpbWUgU3RhbXBpbmcgU2VydmljZXMg
# Q0EgLSBHMjAeFw0xMjEwMTgwMDAwMDBaFw0yMDEyMjkyMzU5NTlaMGIxCzAJBgNV
# BAYTAlVTMR0wGwYDVQQKExRTeW1hbnRlYyBDb3Jwb3JhdGlvbjE0MDIGA1UEAxMr
# U3ltYW50ZWMgVGltZSBTdGFtcGluZyBTZXJ2aWNlcyBTaWduZXIgLSBHNDCCASIw
# DQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAKJjCzlEuLsjp0RJuw7/ofBhClOT
# sJjbrSwPSsVu/4Y8U1UPFc4EPyv9qZaW2b5heQtbyUyGduXgQ0sile7CK0PBn9ho
# tI5AT+6FOLkRxSPyZFjwFTJvTlehroikAtcqHs1L4d1j1ReJMluwXplaqJ0oUA4X
# 7pbbYTtFUR3PElYLkkf8q672Zj1HrHBy55LnX80QucSDZJQZvSWA4ejSIqXQugJ6
# oXeTW2XD7hd0vEGGKtwITIySjJEtnndEH2jWqHR32w5bMotWizO92WPISZ06xcXq
# MwvS8aMb9Iu+2bNXizveBKd6IrIkri7HcMW+ToMmCPsLvalPmQjhEChyqs0CAwEA
# AaOCAVcwggFTMAwGA1UdEwEB/wQCMAAwFgYDVR0lAQH/BAwwCgYIKwYBBQUHAwgw
# DgYDVR0PAQH/BAQDAgeAMHMGCCsGAQUFBwEBBGcwZTAqBggrBgEFBQcwAYYeaHR0
# cDovL3RzLW9jc3Aud3Muc3ltYW50ZWMuY29tMDcGCCsGAQUFBzAChitodHRwOi8v
# dHMtYWlhLndzLnN5bWFudGVjLmNvbS90c3MtY2EtZzIuY2VyMDwGA1UdHwQ1MDMw
# MaAvoC2GK2h0dHA6Ly90cy1jcmwud3Muc3ltYW50ZWMuY29tL3Rzcy1jYS1nMi5j
# cmwwKAYDVR0RBCEwH6QdMBsxGTAXBgNVBAMTEFRpbWVTdGFtcC0yMDQ4LTIwHQYD
# VR0OBBYEFEbGaaMOShQe1UzaUmMXP142vA3mMB8GA1UdIwQYMBaAFF+a9W5czMx0
# mtTdfe8/2+xMgC7dMA0GCSqGSIb3DQEBBQUAA4IBAQB4O7SRKgBM8I9iMDd4o4Qn
# B28Yst4l3KDUlAOqhk4ln5pAAxzdzuN5yyFoBtq2MrRtv/QsJmMz5ElkbQ3mw2cO
# 9wWkNWx8iRbG6bLfsundIMZxD82VdNy2XN69Nx9DeOZ4tc0oBCCjqvFLxIgpkQ6A
# 0RH83Vx2bk9eDkVGQW4NsOo4mrE62glxEPwcebSAe6xp9P2ctgwWK/F/Wwk9m1vi
# FsoTgW0ALjgNqCmPLOGy9FqpAa8VnCwvSRvbIrvD/niUUcOGsYKIXfA9tFGheTMr
# Lnu53CAJE3Hrahlbz+ilMFcsiUk/uc9/yb8+ImhjU5q9aXSsxR08f5Lgw7wc2AR1
# MIIE0DCCA7igAwIBAgIBBzANBgkqhkiG9w0BAQsFADCBgzELMAkGA1UEBhMCVVMx
# EDAOBgNVBAgTB0FyaXpvbmExEzARBgNVBAcTClNjb3R0c2RhbGUxGjAYBgNVBAoT
# EUdvRGFkZHkuY29tLCBJbmMuMTEwLwYDVQQDEyhHbyBEYWRkeSBSb290IENlcnRp
# ZmljYXRlIEF1dGhvcml0eSAtIEcyMB4XDTExMDUwMzA3MDAwMFoXDTMxMDUwMzA3
# MDAwMFowgbQxCzAJBgNVBAYTAlVTMRAwDgYDVQQIEwdBcml6b25hMRMwEQYDVQQH
# EwpTY290dHNkYWxlMRowGAYDVQQKExFHb0RhZGR5LmNvbSwgSW5jLjEtMCsGA1UE
# CxMkaHR0cDovL2NlcnRzLmdvZGFkZHkuY29tL3JlcG9zaXRvcnkvMTMwMQYDVQQD
# EypHbyBEYWRkeSBTZWN1cmUgQ2VydGlmaWNhdGUgQXV0aG9yaXR5IC0gRzIwggEi
# MA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQC54MsQ1K92vdSTYuswZLiBCGzD
# BNliF44v/z5lz4/OYuY8UhzaFkVLVat4a2ODYpDOD2lsmcgaFItMzEUz6ojcnqOv
# K/6AYZ15V8TPLvQ/MDxdR/yaFrzDN5ZBUY4RS1T4KL7QjL7wMDge87Am+GZHY23e
# cSZHjzhHU9FGHbTj3ADqRay9vHHZqm8A29vNMDp5T19MR/gd71vCxJ1gO7GyQ5HY
# pDNO6rPWJ0+tJYqlxvTV0KaudAVkV4i1RFXULSo6Pvi4vekyCgKUZMQWOlDxSq7n
# eTOvDCAHf+jfBDnCaQJsY1L6d8EbyHSHyLmTGFBUNUtpTrw700kuH9zB0lL7AgMB
# AAGjggEaMIIBFjAPBgNVHRMBAf8EBTADAQH/MA4GA1UdDwEB/wQEAwIBBjAdBgNV
# HQ4EFgQUQMK9J47MNIMwojPX+2yz8LQsgM4wHwYDVR0jBBgwFoAUOpqFBxBnKLbv
# 9r0FQW4gwZTaD94wNAYIKwYBBQUHAQEEKDAmMCQGCCsGAQUFBzABhhhodHRwOi8v
# b2NzcC5nb2RhZGR5LmNvbS8wNQYDVR0fBC4wLDAqoCigJoYkaHR0cDovL2NybC5n
# b2RhZGR5LmNvbS9nZHJvb3QtZzIuY3JsMEYGA1UdIAQ/MD0wOwYEVR0gADAzMDEG
# CCsGAQUFBwIBFiVodHRwczovL2NlcnRzLmdvZGFkZHkuY29tL3JlcG9zaXRvcnkv
# MA0GCSqGSIb3DQEBCwUAA4IBAQAIfmyTEMg4uJapkEv/oV9PBO9sPpyIBslQj6Zz
# 91cxG7685C/b+LrTW+C05+Z5Yg4MotdqY3MxtfWoSKQ7CC2iXZDXtHwlTxFWMMS2
# RJ17LJ3lXubvDGGqv+QqG+6EnriDfcFDzkSnE3ANkR/0yBOtg2DZ2HKocyQetawi
# DsoXiWJYRBuriSUBAA/NxBti21G00w9RKpv0vHP8ds42pM3Z2Czqrpv1KrKQ0U11
# GIo/ikGQI31bS/6kA1ibRrLDYGCD+H1QQc7CoZDDu+8CL9IVVO5EFdkKrqeKM+2x
# LXY2JtwE65/3YR8V3Idv7kaWKK2hJn0KCacuBKONvPi8BDABMIIFJjCCBA6gAwIB
# AgIICsxCB6AJxjswDQYJKoZIhvcNAQELBQAwgbQxCzAJBgNVBAYTAlVTMRAwDgYD
# VQQIEwdBcml6b25hMRMwEQYDVQQHEwpTY290dHNkYWxlMRowGAYDVQQKExFHb0Rh
# ZGR5LmNvbSwgSW5jLjEtMCsGA1UECxMkaHR0cDovL2NlcnRzLmdvZGFkZHkuY29t
# L3JlcG9zaXRvcnkvMTMwMQYDVQQDEypHbyBEYWRkeSBTZWN1cmUgQ2VydGlmaWNh
# dGUgQXV0aG9yaXR5IC0gRzIwHhcNMTcwNDA1MjAwMjAwWhcNMjAwNDEwMTk1MTM5
# WjBoMQswCQYDVQQGEwJVUzEOMAwGA1UECBMFVGV4YXMxDzANBgNVBAcTBkF1c3Rp
# bjEbMBkGA1UEChMSQWN0aWFuIENvcnBvcmF0aW9uMRswGQYDVQQDExJBY3RpYW4g
# Q29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDDl5bZ
# gsKcnWzrY8rYrsSmI4l9tC5Ipk1Dp/LHYZQ/1uwgzbiUQavCOKhJ6q+XMGHIOF6P
# N/PkfIfrf1XygR3bJy+f0u4HICioNg0GmfBgEUHVpZTkeRxCT+Dd1FeDDljp9J41
# AVkXh8SnzKvSGKjaUbp3DkH2Y5P24pHR9zhpx2hnqSF0PHOT0fGFcF4SAeYBTIT0
# nSmVPxjs2+NV1M81xtCLLLCF9ceV784DyY7MhoHI9Q7fHpxUJZi0nuHRzyUurofC
# QmJhOLd5vm7re/WUAHfhG3P7dIkm1mI9SHAWid2A/4vGWHb/bR5mwJTDxvPmgfzz
# EVmR8PHPNTsRXZYpAgMBAAGjggGFMIIBgTAMBgNVHRMBAf8EAjAAMBMGA1UdJQQM
# MAoGCCsGAQUFBwMDMA4GA1UdDwEB/wQEAwIHgDA1BgNVHR8ELjAsMCqgKKAmhiRo
# dHRwOi8vY3JsLmdvZGFkZHkuY29tL2dkaWcyczUtMi5jcmwwXQYDVR0gBFYwVDBI
# BgtghkgBhv1tAQcXAjA5MDcGCCsGAQUFBwIBFitodHRwOi8vY2VydGlmaWNhdGVz
# LmdvZGFkZHkuY29tL3JlcG9zaXRvcnkvMAgGBmeBDAEEATB2BggrBgEFBQcBAQRq
# MGgwJAYIKwYBBQUHMAGGGGh0dHA6Ly9vY3NwLmdvZGFkZHkuY29tLzBABggrBgEF
# BQcwAoY0aHR0cDovL2NlcnRpZmljYXRlcy5nb2RhZGR5LmNvbS9yZXBvc2l0b3J5
# L2dkaWcyLmNydDAfBgNVHSMEGDAWgBRAwr0njsw0gzCiM9f7bLPwtCyAzjAdBgNV
# HQ4EFgQUd28uU+LsLPArCXvwS49GX5IaKcUwDQYJKoZIhvcNAQELBQADggEBAKMz
# CNStFW8r/jaaSDXwbWnpmRjeGx3FZbmwH+r2gv2UQ5Se0ckdRm86CfF8AjGnFd/d
# pOSsLWUSWx5+gdqorB9+xY6KrB0br08CfRYvvw+ZYiRrnvLEJyD3hc13tHYpmDOS
# lamg3seFzk/rKe5g/ZXwDQXOiAMEkH5xBHS3BBILQarbiJ14CLM9eLDSeN8L/Gdy
# SvMX4nWy9UusiXP3ZXvuVPUlAOBvDUZaM1BPir0ZWY/m2FOW8bHxXzTG8VtdgY0y
# JIkaii7O3A1kkB9ofB47LQN/CuwcOOKQx0hNL3Nlq7zsFizWk6MO3ZmPNjjimkI6
# mT0NuoY6yd9VefcO95MxggSFMIIEgQIBATCBwTCBtDELMAkGA1UEBhMCVVMxEDAO
# BgNVBAgTB0FyaXpvbmExEzARBgNVBAcTClNjb3R0c2RhbGUxGjAYBgNVBAoTEUdv
# RGFkZHkuY29tLCBJbmMuMS0wKwYDVQQLEyRodHRwOi8vY2VydHMuZ29kYWRkeS5j
# b20vcmVwb3NpdG9yeS8xMzAxBgNVBAMTKkdvIERhZGR5IFNlY3VyZSBDZXJ0aWZp
# Y2F0ZSBBdXRob3JpdHkgLSBHMgIICsxCB6AJxjswCQYFKw4DAhoFAKCBijAZBgkq
# hkiG9w0BCQMxDAYKKwYBBAGCNwIBBDAcBgorBgEEAYI3AgELMQ4wDAYKKwYBBAGC
# NwIBFTAjBgkqhkiG9w0BCQQxFgQUzXrQ5vY1JxjNu/BYxWU3exht7yEwKgYKKwYB
# BAGCNwIBDDEcMBqhGIAWaHR0cDpcXHd3dy5hY3RpYW4uY29tIDANBgkqhkiG9w0B
# AQEFAASCAQAxiqhHko9Jgk74rX3WRK8xEmIpMPJqKQRVYAaLImLmgdjlXNH2cNJE
# Ye59ML+L92/2iF4xdMo13cklg7P0gg8qDlsdLdds94336a15DfkBUEFFuNA0mUTC
# 8oBdZxwkYbuUE9MdV2uvbLjGWk85eGVDkgUGDpOGK2HtqRegEs300QK945f1d+vR
# Tr6Yoo8/gxTHJUSpU6+Pd9KnDZb6f/RI9eI4vJOH7Rw4sJQxMy/Zopc2Q4Et3a/K
# ffiZKkUoGDlkT4qhu3RMuOm40q+GvE2agcm9hi93kTzJszZ76mWtqJYtHwPXBLfZ
# BYsKnBAxgRMo4q9lKPb2Qwf5C0cqASpZoYICCzCCAgcGCSqGSIb3DQEJBjGCAfgw
# ggH0AgEBMHIwXjELMAkGA1UEBhMCVVMxHTAbBgNVBAoTFFN5bWFudGVjIENvcnBv
# cmF0aW9uMTAwLgYDVQQDEydTeW1hbnRlYyBUaW1lIFN0YW1waW5nIFNlcnZpY2Vz
# IENBIC0gRzICEA7P9DjI/r81bgTYapgbGlAwCQYFKw4DAhoFAKBdMBgGCSqGSIb3
# DQEJAzELBgkqhkiG9w0BBwEwHAYJKoZIhvcNAQkFMQ8XDTIwMDMwNzAxMTE1NVow
# IwYJKoZIhvcNAQkEMRYEFIp/wvozMVaTfVGgwtN48KRGW2pCMA0GCSqGSIb3DQEB
# AQUABIIBAEkA6SrshYVneW8kD2/tyNcwiIfs4dh+SJ5gpkW6IocICOpDvVNRKZju
# EUexm79c+x4IKkVaBVkX4YAJqYK5edBxmZbjX2e0EN10e3QMTVsBfJTh4n2wzE5y
# Lj+MVu7DWmN4P/tuOxM5Z5wjYbmNuUWOMsDQoWwrxrVwSTfZPuZ+tpPPdPCLK+IX
# f/20bzdrx6jU0RYUNrhLFgv9ODJcwdi5zhDzM/+OxpRA0aO9s5ksCUKMrtRwiv/y
# tU6c7ydZi0Q6fbW7BCS8Rb2eOXdtP6X8MO3HXKPgM9nUDC4StNFyGtMirzQckUjV
# HW6ePuAQXAvSAWCBTl8L84/tyPVYitU=
# SIG # End signature block
